@@ -16,14 +16,15 @@
 # specific language governing permissions and limitations
 # under the License.
 #
-import os
-import re
-import pexpect
-import stat
-import shlex
-import tempfile
 import logging
+import os
+import pexpect
+import re
+import shlex
+import stat
+import tempfile
 
+import six
 from uuid import uuid4
 try:
     from collections.abc import Mapping
@@ -32,7 +33,7 @@ except ImportError:
 
 from distutils.dir_util import copy_tree
 
-from six import iteritems, string_types
+from six import iteritems, string_types, text_type
 
 from ansible_runner import output
 from ansible_runner.exceptions import ConfigurationError
@@ -77,10 +78,14 @@ class RunnerConfig(object):
                  rotate_artifacts=0, host_pattern=None, binary=None, extravars=None, suppress_ansible_output=False,
                  process_isolation=False, process_isolation_executable=None, process_isolation_path=None,
                  process_isolation_hide_paths=None, process_isolation_show_paths=None, process_isolation_ro_paths=None,
-                 tags=None, skip_tags=None, fact_cache_type='jsonfile', fact_cache=None, project_dir=None,
-                 directory_isolation_base_path=None):
+                 resource_profiling=False, resource_profiling_base_cgroup='ansible-runner', resource_profiling_cpu_poll_interval=0.25,
+                 resource_profiling_memory_poll_interval=0.25, resource_profiling_pid_poll_interval=0.25,
+                 resource_profiling_results_dir=None,
+                 tags=None, skip_tags=None, fact_cache_type='jsonfile', fact_cache=None, ssh_key=None,
+                 project_dir=None, directory_isolation_base_path=None, envvars=None, forks=None, cmdline=None, omit_event_data=False,
+                 only_failed_event_data=False):
         self.private_data_dir = os.path.abspath(private_data_dir)
-        self.ident = ident
+        self.ident = str(ident)
         self.json_mode = json_mode
         self.playbook = playbook
         self.inventory = inventory
@@ -92,18 +97,30 @@ class RunnerConfig(object):
         self.binary = binary
         self.rotate_artifacts = rotate_artifacts
         self.artifact_dir = os.path.abspath(artifact_dir or self.private_data_dir)
-        if self.ident is None:
-            self.artifact_dir = os.path.join(self.artifact_dir, "artifacts")
+
+        if artifact_dir is None:
+            self.artifact_dir = os.path.join(self.private_data_dir, 'artifacts')
         else:
-            self.artifact_dir = os.path.join(self.artifact_dir, "artifacts", "{}".format(self.ident))
+            self.artifact_dir = os.path.abspath(artifact_dir)
+
+        if self.ident is not None:
+            self.artifact_dir = os.path.join(self.artifact_dir, "{}".format(self.ident))
 
         self.extra_vars = extravars
         self.process_isolation = process_isolation
         self.process_isolation_executable = process_isolation_executable
         self.process_isolation_path = process_isolation_path
+        self.process_isolation_path_actual = None
         self.process_isolation_hide_paths = process_isolation_hide_paths
         self.process_isolation_show_paths = process_isolation_show_paths
         self.process_isolation_ro_paths = process_isolation_ro_paths
+        self.resource_profiling = resource_profiling
+        self.resource_profiling_base_cgroup = resource_profiling_base_cgroup
+        self.resource_profiling_cpu_poll_interval = resource_profiling_cpu_poll_interval
+        self.resource_profiling_memory_poll_interval = resource_profiling_memory_poll_interval
+        self.resource_profiling_pid_poll_interval = resource_profiling_pid_poll_interval
+        self.resource_profiling_results_dir = resource_profiling_results_dir
+
         self.directory_isolation_path = directory_isolation_base_path
         if not project_dir:
             self.project_dir = os.path.join(self.private_data_dir, 'project')
@@ -117,7 +134,13 @@ class RunnerConfig(object):
         self.skip_tags = skip_tags
         self.fact_cache_type = fact_cache_type
         self.fact_cache = os.path.join(self.artifact_dir, fact_cache or 'fact_cache') if self.fact_cache_type == 'jsonfile' else None
+        self.ssh_key_data = ssh_key
         self.execution_mode = ExecutionMode.NONE
+        self.envvars = envvars
+        self.forks = forks
+        self.cmdline_args = cmdline
+        self.omit_event_data = omit_event_data
+        self.only_failed_event_data = only_failed_event_data
 
     def prepare(self):
         """
@@ -171,15 +194,38 @@ class RunnerConfig(object):
         python_path = self.env.get('PYTHONPATH', os.getenv('PYTHONPATH', ''))
         if python_path and not python_path.endswith(':'):
             python_path += ':'
-        self.env['ANSIBLE_CALLBACK_PLUGINS'] = callback_dir
+        self.env['ANSIBLE_CALLBACK_PLUGINS'] = ':'.join(filter(None,(self.env.get('ANSIBLE_CALLBACK_PLUGINS'), callback_dir)))
         if 'AD_HOC_COMMAND_ID' in self.env:
             self.env['ANSIBLE_STDOUT_CALLBACK'] = 'minimal'
         else:
             self.env['ANSIBLE_STDOUT_CALLBACK'] = 'awx_display'
         self.env['ANSIBLE_RETRY_FILES_ENABLED'] = 'False'
-        self.env['ANSIBLE_HOST_KEY_CHECKING'] = 'False'
+        if 'ANSIBLE_HOST_KEY_CHECKING' not in self.env:
+            self.env['ANSIBLE_HOST_KEY_CHECKING'] = 'False'
         self.env['AWX_ISOLATED_DATA_DIR'] = self.artifact_dir
 
+        if self.resource_profiling:
+            callback_whitelist = os.environ.get('ANSIBLE_CALLBACK_WHITELIST', '').strip()
+            self.env['ANSIBLE_CALLBACK_WHITELIST'] = ','.join(filter(None, [callback_whitelist, 'cgroup_perf_recap']))
+            self.env['CGROUP_CONTROL_GROUP'] = '{}/{}'.format(self.resource_profiling_base_cgroup, self.ident)
+            if self.resource_profiling_results_dir:
+                cgroup_output_dir = self.resource_profiling_results_dir
+            else:
+                cgroup_output_dir = os.path.normpath(os.path.join(self.private_data_dir, 'profiling_data'))
+
+            # Create results directory if it does not exist
+            if not os.path.isdir(cgroup_output_dir):
+                os.mkdir(cgroup_output_dir, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)
+
+            self.env['CGROUP_OUTPUT_DIR'] = cgroup_output_dir
+            self.env['CGROUP_OUTPUT_FORMAT'] = 'json'
+            self.env['CGROUP_CPU_POLL_INTERVAL'] = str(self.resource_profiling_cpu_poll_interval)
+            self.env['CGROUP_MEMORY_POLL_INTERVAL'] = str(self.resource_profiling_memory_poll_interval)
+            self.env['CGROUP_PID_POLL_INTERVAL'] = str(self.resource_profiling_pid_poll_interval)
+            self.env['CGROUP_FILE_PER_TASK'] = 'True'
+            self.env['CGROUP_WRITE_FILES'] = 'True'
+            self.env['CGROUP_DISPLAY_RECAP'] = 'False'
+
         self.env['PYTHONPATH'] = python_path + callback_dir
         if self.roles_path:
             self.env['ANSIBLE_ROLES_PATH'] = ':'.join(self.roles_path)
@@ -187,15 +233,21 @@ class RunnerConfig(object):
         if self.process_isolation:
             self.command = self.wrap_args_with_process_isolation(self.command)
 
+        if self.resource_profiling and self.execution_mode == ExecutionMode.ANSIBLE_PLAYBOOK:
+            self.command = self.wrap_args_with_cgexec(self.command)
+
         if self.fact_cache_type == 'jsonfile':
             self.env['ANSIBLE_CACHE_PLUGIN'] = 'jsonfile'
             self.env['ANSIBLE_CACHE_PLUGIN_CONNECTION'] = self.fact_cache
 
+        self.env["RUNNER_OMIT_EVENTS"] = str(self.omit_event_data)
+        self.env["RUNNER_ONLY_FAILED_EVENTS"] = str(self.only_failed_event_data)
+
     def prepare_inventory(self):
         """
         Prepares the inventory default under ``private_data_dir`` if it's not overridden by the constructor.
         """
-        if self.inventory is None:
+        if self.inventory is None and os.path.exists(os.path.join(self.private_data_dir, "inventory")):
             self.inventory  = os.path.join(self.private_data_dir, "inventory")
 
     def prepare_env(self):
@@ -215,16 +267,20 @@ class RunnerConfig(object):
         self.expect_passwords[pexpect.TIMEOUT] = None
         self.expect_passwords[pexpect.EOF] = None
 
+        # seed env with existing shell env
+        self.env = os.environ.copy()
+        if self.envvars and isinstance(self.envvars, dict):
+            if six.PY2:
+                self.env.update({k.decode('utf-8'):v.decode('utf-8') for k, v in self.envvars.items()})
+            else:
+                self.env.update({k:v for k, v in self.envvars.items()})
         try:
-            # seed env with existing shell env
-            self.env = os.environ.copy()
             envvars = self.loader.load_file('env/envvars', Mapping)
             if envvars:
-                self.env.update({k:str(v) for k, v in envvars.items()})
+                self.env.update({six.text_type(k):six.text_type(v) for k, v in envvars.items()})
         except ConfigurationError:
             output.debug("Not loading environment vars")
             # Still need to pass default environment to pexpect
-            self.env = os.environ.copy()
 
         try:
             self.settings = self.loader.load_file('env/settings', Mapping)
@@ -233,7 +289,8 @@ class RunnerConfig(object):
             self.settings = dict()
 
         try:
-            self.ssh_key_data = self.loader.load_file('env/ssh_key', string_types)
+            if self.ssh_key_data is None:
+                self.ssh_key_data = self.loader.load_file('env/ssh_key', string_types)
         except ConfigurationError:
             output.debug("Not loading ssh key")
             self.ssh_key_data = None
@@ -248,7 +305,13 @@ class RunnerConfig(object):
         self.process_isolation_hide_paths = self.settings.get('process_isolation_hide_paths', self.process_isolation_hide_paths)
         self.process_isolation_show_paths = self.settings.get('process_isolation_show_paths', self.process_isolation_show_paths)
         self.process_isolation_ro_paths = self.settings.get('process_isolation_ro_paths', self.process_isolation_ro_paths)
-
+        self.resource_profiling = self.settings.get('resource_profiling', self.resource_profiling)
+        self.resource_profiling_base_cgroup = self.settings.get('resource_profiling_base_cgroup', self.resource_profiling_base_cgroup)
+        self.resource_profiling_cpu_poll_interval = self.settings.get('resource_profiling_cpu_poll_interval', self.resource_profiling_cpu_poll_interval)
+        self.resource_profiling_memory_poll_interval = self.settings.get('resource_profiling_memory_poll_interval',
+                                                                         self.resource_profiling_memory_poll_interval)
+        self.resource_profiling_pid_poll_interval = self.settings.get('resource_profiling_pid_poll_interval', self.resource_profiling_pid_poll_interval)
+        self.resource_profiling_results_dir = self.settings.get('resource_profiling_results_dir', self.resource_profiling_results_dir)
         self.pexpect_use_poll = self.settings.get('pexpect_use_poll', True)
         self.suppress_ansible_output = self.settings.get('suppress_ansible_output', self.quiet)
         self.directory_isolation_cleanup = bool(self.settings.get('directory_isolation_cleanup', True))
@@ -274,8 +337,10 @@ class RunnerConfig(object):
         and if not calls :py:meth:`ansible_runner.runner_config.RunnerConfig.generate_ansible_command`
         """
         try:
-            cmdline_args = self.loader.load_file('args', string_types)
-            self.command = shlex.split(cmdline_args.decode('utf-8'))
+            cmdline_args = self.loader.load_file('args', string_types, encoding=None)
+            if six.PY2 and isinstance(cmdline_args, text_type):
+                cmdline_args = cmdline_args.encode('utf-8')
+            self.command = shlex.split(cmdline_args)
             self.execution_mode = ExecutionMode.RAW
         except ConfigurationError:
             self.command = self.generate_ansible_command()
@@ -299,13 +364,22 @@ class RunnerConfig(object):
         exec_list = [base_command]
 
         try:
-            cmdline_args = self.loader.load_file('env/cmdline', string_types)
-            args = shlex.split(cmdline_args.decode('utf-8'))
+            if self.cmdline_args:
+                cmdline_args = self.cmdline_args
+            else:
+                cmdline_args = self.loader.load_file('env/cmdline', string_types, encoding=None)
+
+            if six.PY2 and isinstance(cmdline_args, text_type):
+                cmdline_args = cmdline_args.encode('utf-8')
+
+            args = shlex.split(cmdline_args)
             exec_list.extend(args)
         except ConfigurationError:
             pass
 
-        if isinstance(self.inventory, list):
+        if self.inventory is None:
+            pass
+        elif isinstance(self.inventory, list):
             for i in self.inventory:
                 exec_list.append("-i")
                 exec_list.append(i)
@@ -319,24 +393,36 @@ class RunnerConfig(object):
 
         if self.loader.isfile('env/extravars'):
             exec_list.extend(['-e', '@{}'.format(self.loader.abspath('env/extravars'))])
-        if isinstance(self.extra_vars, dict) and self.extra_vars:
-            exec_list.extend(
-                [
-                    '-e',
-                    '\'%s\'' % ' '.join(
-                        ["{}=\"{}\"".format(k, self.extra_vars[k]) for k in self.extra_vars]
-                    )
-                ]
-            )
+
+        if self.extra_vars:
+            if isinstance(self.extra_vars, dict) and self.extra_vars:
+                extra_vars_list = []
+                for k in self.extra_vars:
+                    if isinstance(self.extra_vars[k], str):
+                        extra_vars_list.append("\"{}\":\"{}\"".format(k, self.extra_vars[k]))
+                    else:
+                        extra_vars_list.append("\"{}\":{}".format(k, self.extra_vars[k]))
+                exec_list.extend(
+                    [
+                        '-e',
+                        '{%s}' % ','.join(extra_vars_list)
+                    ]
+                )
+            elif self.loader.isfile(self.extra_vars):
+                exec_list.extend(['-e', '@{}'.format(self.loader.abspath(self.extra_vars))])
+
         if self.verbosity:
             v = 'v' * self.verbosity
-            exec_list.append('-%s' % v)
+            exec_list.append('-{}'.format(v))
 
         if self.tags:
-            exec_list.extend(['--tags', '%s' % self.tags])
+            exec_list.extend(['--tags', '{}'.format(self.tags)])
 
         if self.skip_tags:
-            exec_list.extend(['--skip-tags', '%s' % self.skip_tags])
+            exec_list.extend(['--skip-tags', '{}'.format(self.skip_tags)])
+
+        if self.forks:
+            exec_list.extend(['--forks', '{}'.format(self.forks)])
 
         # Other parameters
         if self.execution_mode == ExecutionMode.ANSIBLE_PLAYBOOK:
@@ -362,14 +448,23 @@ class RunnerConfig(object):
         os.chmod(path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
         return path
 
+    def wrap_args_with_cgexec(self, args):
+        '''
+        Wrap existing command line with cgexec in order to profile resource usage
+        '''
+        new_args = ['cgexec', '--sticky', '-g', 'cpuacct,memory,pids:{}/{}'.format(self.resource_profiling_base_cgroup, self.ident)]
+        new_args.extend(args)
+        return new_args
+
+
     def wrap_args_with_process_isolation(self, args):
         '''
         Wrap existing command line with bwrap to restrict access to:
          - self.process_isolation_path (generally, /tmp) (except for own /tmp files)
         '''
         cwd = os.path.realpath(self.cwd)
-        pi_temp_dir = self.build_process_isolation_temp_dir()
-        new_args = [self.process_isolation_executable or 'bwrap', '--unshare-pid', '--dev-bind', '/', '/', '--proc', '/proc']
+        self.process_isolation_path_actual = self.build_process_isolation_temp_dir()
+        new_args = [self.process_isolation_executable or 'bwrap', '--die-with-parent', '--unshare-pid', '--dev-bind', '/', '/', '--proc', '/proc']
 
         for path in sorted(set(self.process_isolation_hide_paths or [])):
             if not os.path.exists(path):
@@ -377,10 +472,10 @@ class RunnerConfig(object):
                 continue
             path = os.path.realpath(path)
             if os.path.isdir(path):
-                new_path = tempfile.mkdtemp(dir=pi_temp_dir)
+                new_path = tempfile.mkdtemp(dir=self.process_isolation_path_actual)
                 os.chmod(new_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
             else:
-                handle, new_path = tempfile.mkstemp(dir=pi_temp_dir)
+                handle, new_path = tempfile.mkstemp(dir=self.process_isolation_path_actual)
                 os.close(handle)
                 os.chmod(new_path, stat.S_IRUSR | stat.S_IWUSR)
             new_args.extend(['--bind', '{0}'.format(new_path), '{0}'.format(path)])
@@ -410,7 +505,7 @@ class RunnerConfig(object):
             if self.directory_isolation_path is not None:
                 new_args.extend(['--chdir', os.path.realpath(self.directory_isolation_path)])
             else:
-                new_args.extend(['--chdir', self.project_dir])
+                new_args.extend(['--chdir', os.path.realpath(self.project_dir)])
         elif self.execution_mode == ExecutionMode.ANSIBLE:
             # ad-hoc runs should cwd to the root of the private data dir
             new_args.extend(['--chdir', os.path.realpath(self.private_data_dir)])
